import 'dart:convert';
import 'dart:ffi';

import 'package:flutter/services.dart';

import 'kameleoon_data.dart';
import 'kameleoon_exception.dart';

class KameleoonClientCallback {
  late Function readyCallback;
  Function? failCallback;
  int? expires;

  KameleoonClientCallback(
      Function readyCallback, Function failCallback, int expires) {
    this.readyCallback = readyCallback;
    this.failCallback = failCallback;
    this.expires = expires;
  }
}

class KameleoonClient {
  static const MethodChannel _channel =
      const MethodChannel("kameleoon_client_flutter");

  bool? ready = false;
  List<KameleoonClientCallback> callbacks = [];
  String? siteCode;

  KameleoonClient(String siteCode) {
    this.siteCode = siteCode;
    _channel.setMethodCallHandler(nativeMethodCallHandler);
    _channel.invokeMethod("initializeKameleoonClient",
        <String, dynamic>{"siteCode": this.siteCode});
  }


  // To track conversion, use the trackConversion() method. This method requires visitorCode and goalID to track conversion on this particular goal. In addition, this method also accepts revenue as a third optional argument to track revenue. The visitorCode should be identical to the one that was used when triggering the experiment.
  // The trackConversion() method doesn't return any value. This method is non-blocking as the server call is made asynchronously.
  Future<void> trackConversion(String visitorCode, int goalID,
      [Float? revenue]) async {
    await _channel.invokeMethod("trackConversion", <String, dynamic>{
      "visitorCode": visitorCode,
      "goalID": goalID,
      "revenue": revenue,
      "siteCode": this.siteCode
    });
  }

  // Running an A/B experiment on your Flutter application means bucketing your users into several groups (one per variation). The SDK takes care of this bucketing (and the associated reporting) automatically.
  // Triggering an experiment by calling the triggerExperiment() method will register a random variation for a given visitorCode. If this visitorCode is already associated with a variation (most likely the user had already been exposed to the experiment previously), then it will return the previous variation associated with a given experiment.
  Future<int?> triggerExperiment(String visitorCode, int experimentID) async {
    try {
      return await _channel.invokeMethod("triggerExperiment", <String, dynamic>{
        "visitorCode": visitorCode,
        "experimentID": experimentID,
        "siteCode": this.siteCode
      });
    } on PlatformException catch (e) {
      switch (e.code) {
        case "SDKNotReady":
          throw SDKNotReady();
        case "ExperimentConfigurationNotFound":
          throw ExperimentConfigurationNotFound();
        case "NotTargeted":
          throw NotTargeted();
        case "NotActivated":
          throw NotActivated();
        case "VisitorCodeNotValid":
          throw VisitorCodeNotValid(e.message);
        default:
          throw Exception(e);
      }
    }
  }

  // To activate a feature toggle, call the activateFeature() method of our SDK.
  // This method takes a visitorCode and featureKey (or featureID) as mandatory arguments to check if the specified feature will be active for a given user.
  // If such a user has never been associated with this feature flag, the SDK returns a boolean value randomly (true if the user should have this feature or false if not). If a user with a given visitorCode is already registered with this feature flag, it will detect the previous featureFlag value.
  // You have to make sure that proper error handling is set up in your code as shown in the example to the right to catch potential exceptions.
  Future<bool?> activateFeature(String visitorCode, var featureKeyOrID) async {
    try {
      if (featureKeyOrID is int) {
        return await _channel.invokeMethod("activateFeature", <String, dynamic>{
          "visitorCode": visitorCode,
          "featureID": featureKeyOrID,
          "siteCode": this.siteCode
        });
      } else {
        return await _channel.invokeMethod("activateFeature", <String, dynamic>{
          "visitorCode": visitorCode,
          "featureKey": featureKeyOrID,
          "siteCode": this.siteCode
        });
      }
    } on PlatformException catch (e) {
      switch (e.code) {
        case "SDKNotReady":
          throw SDKNotReady();
        case "FeatureConfigurationNotFound":
          throw FeatureConfigurationNotFound();
        case "NotTargeted":
          throw NotTargeted();
        case "NotActivated":
          throw NotActivated();
        case "VisitorCodeNotValid":
          throw VisitorCodeNotValid(e.message);
        default:
          throw Exception(e);
      }
    }
  }

  Future<String?> obtainVisitorCode([String? defaultVisitorCode]) async {
    return await _channel.invokeMethod("obtainVisitorCode", <String, dynamic>{
      "defaultVisitorCode": defaultVisitorCode,
      "siteCode": this.siteCode
    });
  }

  // To retrieve JSON data associated with a variation, call the obtainVariationAssociatedData() method of our SDK. The JSON data usually represents some metadata of the variation, and can be configured on our web application interface or via our Automation API.
  // This method takes the variationID as a parameter and will return the data as a JSON object. It will throw an exception (KameleoonException.VariationConfigurationNotFound) if the variation ID is wrong or corresponds to an experiment that is not yet online.
  Future<dynamic> obtainVariationAssociatedData(int variationID) async {
    try {
      String objAsString = await (_channel.invokeMethod(
          "obtainVariationAssociatedData", <String, dynamic>{
        "variationID": variationID,
        "siteCode": this.siteCode
      }));
      return jsonDecode(objAsString);
    } on PlatformException catch (e) {
      switch (e.code) {
        case "VariationConfigurationNotFound":
          throw VariationConfigurationNotFound();
        default:
          throw Exception(e);
      }
    }
  }

  // To retrieve a feature variable, call the obtainFeatureVariable() method of our SDK. A feature variable can be changed easily via our web application.
  // This method takes two input parameters: featureKey and variableKey. It will return the data as an Object instance. Usually it should be casted to the expected type (the one defined on the web interface). It will throw an exception (KameleoonException.FeatureConfigurationNotFound) if the requested feature has not been found in the internal configuration of the SDK.
  Future<Object?> obtainFeatureVariable(
      var featureKeyOrID, String variableKey) async {
    if (featureKeyOrID is int) {
      return await _channel
          .invokeMethod("obtainFeatureVariable", <String, dynamic>{
        "variableKey": variableKey,
        "featureID": featureKeyOrID,
        "siteCode": this.siteCode
      });
    } else {
      return await _channel
          .invokeMethod("obtainFeatureVariable", <String, dynamic>{
        "variableKey": variableKey,
        "featureKey": featureKeyOrID,
        "siteCode": this.siteCode
      });
    }
  }

  // For mobile SDKs, the initialization of the Kameleoon Client is not immediate, as it needs to perform a server call to retrieve the current configuration for all active experiments. It is recommended to check if the SDK is ready by calling this method before triggering an experiment. Alternatively, you could setup exception catching around the triggerExperiment() method to look for the KameleoonException.SDKNotReady exception, or you could use the runWhenReady() method with a callback as detailed in the next paragraph.
  bool? isReady() {
    return this.ready;
  }

  // To associate various data with the current user, we can use the addData() method. This method requires the visitorCode as a first parameter, and then accept several additional parameters. Those additional parameters represent the various Data Types allowed in Kameleoon.
  // Note that the addData() method doesn't return any value and doesn't interact with the Kameleoon back-end servers by itself. Instead, every data declared is saved for further sending via the flush() method described in the next paragraph. This allows a minimal number of server calls to be made, as data are usually regrouped in a single server call triggered by the execution of flush().
  Future<void> addData(String visitorCode, Data data) async {
    var parameters = <String, dynamic>{};
    parameters.putIfAbsent("visitorCode", () => visitorCode);
    parameters.putIfAbsent("dataType", () => data.dataType);
    parameters.putIfAbsent("siteCode", () => this.siteCode);
    switch (data.dataType) {
      case "CustomData":
        CustomData d = data as CustomData;
        parameters.putIfAbsent("id", () => d.id);
        parameters.putIfAbsent("value", () => d.value);
        break;
      case "PageView":
        PageView d = data as PageView;
        parameters.putIfAbsent("url", () => d.url);
        parameters.putIfAbsent("title", () => d.title);
        parameters.putIfAbsent("referrer", () => d.referrer);
        break;
      case "Interest":
        Interest d = data as Interest;
        parameters.putIfAbsent("index", () => d.index);
        break;
      case "Conversion":
        Conversion d = data as Conversion;
        parameters.putIfAbsent("goalId", () => d.goalId);
        parameters.putIfAbsent("revenue", () => d.revenue);
        parameters.putIfAbsent("negative", () => d.negative);
        break;
      case "Browser":
        Browser d = data as Browser;
        parameters.putIfAbsent("browser", () => d.browser);
        break;
    }
    await _channel.invokeMethod("addData", parameters);
  }

  // For mobile SDKs, the initialization of the Kameleoon Client is not immediate, as it needs to perform a server call to retrieve the current configuration for all active experiments. The runWhenReady() method of the KameleoonClient class allows to pass a callback that will be executed as soon as the SDK is ready for use. It also allows the use of a timeout.
  // The first callback will be executed once the Kameleoon client is ready, and should contain code triggering an experiment and implementing variations. The second callback will be executed if the specified timeout happens before the client is initialized. Usually this case should implement the "reference" variation, as the user will be "out of the experiment" if a timeout takes place.
  void runWhenReady(Function readyCallback, Function failCallback,
      [int timeout = 2000]) async {
    if (this.ready!) {
      await readyCallback.call();
    } else {
      this
          .callbacks
          .add(new KameleoonClientCallback(readyCallback, failCallback, timeout));
      await _channel.invokeMethod("runWhenReady",
          <String, dynamic>{"timeout": timeout, "siteCode": this.siteCode});
    }
  }

  // Data associated with the current user via addData() method is not sent immediately to the server. It is stored and accumulated until it is sent automatically by the triggerExperiment() or trackConversion() methods, or manually by the flush() method. This allows the developer to exactly control when the data is flushed to our servers. For instance, if you call the addData() method a dozen times, it would be a waste of ressources to send data to the server after each addData() invocation. Just call flush() once at the end.
  // The flush() method doesn't return any value. This method is non-blocking as the server call is made asynchronously.
  void flush([String? visitorCode]) async {
    if (visitorCode != null) {
      await _channel.invokeMethod("flush", <String, dynamic>{
        "visitorCode": visitorCode,
      });
    } else {
      await _channel.invokeMethod("flush");
    }
  }

  void callbackAfterStatusChange(bool? isReady,
      [KameleoonClientCallback? callback]) {
    List<int> idsToDelete = [];
    for (int i = 0; i < this.callbacks.length; i++) {
      if (isReady!) {
        this.callbacks[i].readyCallback();
        idsToDelete.add(i);
      }
    }

    for (int i = idsToDelete.length - 1; i >= 0; i--) {
      this.callbacks.removeAt(i);
    }
  }

  Future<void> nativeMethodCallHandler(MethodCall call) async {
    switch (call.method) {
      case "setReady":
        this.ready = call.arguments;
        callbackAfterStatusChange(this.ready);
        break;
      default:
        break;
    }
  }
}
